from shaped_bloom_filter.exceptions import (
    BloomFilterIncorrectConstructorValues,
    BloomFilterIncorrectInputException,
)
from shaped_bloom_filter.filter import BloomFilter, BloomFilterExtended



import glob
import os

import cffi
import pkg_resources


def _find_libbloomf_dll_paths():
    """
    Finds the libbloomf dynamic library file.
    """
    here = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))

    def make_paths(ext):
        return [
            *glob.glob(os.path.join(here, "libbloomf*.{}".format(ext))),
            *glob.glob(os.path.join(here, "..", "libbloomf*.{}".format(ext))),
            *glob.glob(os.path.join(here, "../..", "libbloomf*.{}".format(ext))),
        ]

    return make_paths("so") + make_paths("dll") + make_paths("pyd")


def _get_libbloomf_header_file():
    """
    Retrieve the preamble header file necessary for the bloom dynamic library.
    """
    pkg_name = __name__.split(".")[0]
    return pkg_resources.resource_filename(pkg_name, "golib/cdef.h")


def load_bloom_filter_dll():
    """
    Loads and returns an ffi context and the bloom filter dynamic library file.
    """
    ffi = cffi.FFI()

    dll_paths = _find_libbloomf_dll_paths()
    if not dll_paths:
        raise RuntimeError("Can't find any libbloomf dynamic library files")

    with open(_get_libbloomf_header_file(), "r") as f:
        file_data = f.read()
        ffi.cdef(file_data)

    lib = ffi.dlopen(dll_paths[0])
    return ffi, lib



import random

import pytest
from shaped_bloom_filter import BloomFilter, BloomFilterExtended
from shaped_bloom_filter.exceptions import (
    BloomFilterIncorrectConstructorValues,
    BloomFilterIncorrectInputException,
)


def default_parametrized_bloom_filter(max_elements=10, error_rate=0.01):
    return BloomFilter(
        max_elements=max_elements,
        error_rate=error_rate,
    )


def default_parametrized_bloom_filter_extended():
    return BloomFilterExtended(
        max_elements=10,
        error_rate=0.01,
    )


def test_bloom_filter_creation_no_args():
    with pytest.raises(BloomFilterIncorrectConstructorValues):
        BloomFilter()


def test_bloom_filter_creation_all_args():
    with pytest.raises(BloomFilterIncorrectConstructorValues):
        BloomFilter(
            max_elements=10,
            error_rate=0.01,
            restore_from_serialized=bytes([0, 1, 2, 3]),
        )


def test_bloom_filter_creation_params():
    default_parametrized_bloom_filter()


def test_bloom_filter_creation_serialized():
    bf = default_parametrized_bloom_filter()
    serialized = bf.serialize()

    new_bf = BloomFilter(restore_from_serialized=serialized)
    new_serialized = new_bf.serialize()

    assert serialized == new_serialized


def test_bloom_filter_is_member():
    bf = default_parametrized_bloom_filter()
    assert bf.is_member(3) == False
    bf.add(3)
    assert bf.is_member(3) == True


def test_bloom_filter_are_members():
    bf = default_parametrized_bloom_filter()
    assert any(bf.are_members(list(range(10)))) == False
    bf.add_batch([1, 5, 6])
    assert bf.are_members(list(range(10))) == [0, 1, 0, 0, 0, 1, 1, 0, 0, 0]


def test_bloom_filter_serialized_members():
    bf = default_parametrized_bloom_filter()
    bf.add_batch([1, 5, 6])
    serialized = bf.serialize()

    new_bf = BloomFilter(restore_from_serialized=serialized)
    assert new_bf.are_members(list(range(10))) == [0, 1, 0, 0, 0, 1, 1, 0, 0, 0]


def test_bloom_filter_from_loaded_filter():
    bf = default_parametrized_bloom_filter()
    bf.add_batch([1, 5, 6])
    serialized = bf.serialize()

    new_bf = default_parametrized_bloom_filter()
    new_bf.load(serialized)

    assert new_bf.are_members(list(range(10))) == [0, 1, 0, 0, 0, 1, 1, 0, 0, 0]


def test_bloom_filter_extended_add_one_member_bad_input():
    bf = default_parametrized_bloom_filter_extended()
    with pytest.raises(BloomFilterIncorrectInputException):
        bf.add_one_member("some random string")


def test_bloom_filter_extended_add_one_member_overflow_error():
    bf = default_parametrized_bloom_filter_extended()
    with pytest.raises(OverflowError):
        bf.add_one_member(256)
    with pytest.raises(OverflowError):
        bf.add_one_member([256])


def test_bloom_filter_extended_is_one_member():
    bf = default_parametrized_bloom_filter_extended()
    assert bf.is_one_member([0, 0, 0]) == False
    bf.add_one_member([0, 5, 128])
    assert bf.is_one_member([0, 5, 128]) == True


def test_bloom_filter_extended_multiple_addition_asserts():
    bf = default_parametrized_bloom_filter_extended()
    members_to_add = [0, 5, 9]
    members_to_test = list(range(10))

    for member in members_to_add:
        bf.add_one_member(member)

    for member in members_to_test:
        result = bf.is_one_member(member)
        if member in members_to_add:
            assert result == True, f"member {member} not found to be set"
        else:
            assert result == False, f"member {member} was set"



import pickle
from typing import List, Optional, Union

from shaped_bloom_filter.binding import load_bloom_filter_dll
from shaped_bloom_filter.exceptions import (
    BloomFilterIncorrectConstructorValues,
    BloomFilterIncorrectInputException,
)


class BloomFilter:
    def __init__(
        self,
        max_elements: Optional[int] = None,
        error_rate: Optional[float] = None,
        restore_from_serialized: Optional[bytes] = None,
    ):
        self._ffi, self._libbloomf = load_bloom_filter_dll()
        if max_elements and error_rate and restore_from_serialized:
            raise BloomFilterIncorrectConstructorValues(
                "either set max_elements and error_rate or set restore_from_serialized params"
            )
        if max_elements is not None and error_rate is not None:
            self._filter = self._libbloomf.NewWithEstimates(max_elements, error_rate)
        elif restore_from_serialized is not None:
            self.load(restore_from_serialized)
        else:
            raise BloomFilterIncorrectConstructorValues(
                "either set max_elements and error_rate or set restore_from_serialized params"
            )

    def add(self, var: int):
        """
        Add a single 32-bit integer key to the filter.
        """
        self.add_batch([var])

    def add_batch(self, var: List[int]):
        """
        Add a list of 32-bit integer keys to the filter.
        """
        data = self._ffi.new("GoUint[]", var)
        length = len(data)

        go_slice = self._ffi.new(
            "GoSlice*",
            {
                "data": self._ffi.cast("void*", data),
                "len": length,
                "cap": length,
            },
        )

        self._libbloomf.AddListUint(self._filter, go_slice[0])

    def is_member(self, var: int) -> bool:
        """
        Check if a given 32-bit integer key has been set.
        """
        return self.are_members([var])[0]

    def are_members(self, var: List[int]) -> List[bool]:
        """
        Check if a given list of 32-bit integer keys have been set.
        A boolean list is returned.
        """
        data = self._ffi.new("GoUint[]", var)
        length = len(var)

        go_slice = self._ffi.new(
            "GoSlice*",
            {
                "data": self._ffi.cast("void*", data),
                "len": length,
                "cap": length,
            },
        )

        out = self._libbloomf.TestListUint(self._filter, go_slice[0])
        try:
            unpacked = list(self._ffi.unpack(out, length))
        finally:
            self._libbloomf.free(self._ffi.cast("void*", out))
        return unpacked

    def serialize(self) -> bytes:
        """
        Serialize the filter for storing purposes.
        To restore it, pass the returned bytes into the constructor's restore_from_serialized parameter.
        """
        serializable = {
            "m": self._filter.m,
            "k": self._filter.k,
            "b_length": self._filter.b_length,
            "b": self._ffi.unpack(self._filter.b, self._filter.b_length),
        }
        return pickle.dumps(serializable, protocol=pickle.HIGHEST_PROTOCOL)

    def load(self, serialized_bloom_filter: bytes):
        """
        Restore a serialized bloom filter into the existing context.
        Computationally faster than creating a new filter from scratch.
        """
        self.__del__()
        deserialized = pickle.loads(serialized_bloom_filter)

        m = deserialized["m"]
        k = deserialized["k"]
        b_length = deserialized["b_length"]
        b = deserialized["b"]

        data = self._ffi.new("GoUint8[]", list(b))
        go_slice = self._ffi.new(
            "GoSlice*",
            {
                "data": self._ffi.cast("void*", data),
                "len": b_length,
                "cap": b_length,
            },
        )
        self._filter = self._libbloomf.NewFromSerialized(m, k, b_length, go_slice[0])

    def __contains__(self, var: int) -> bool:
        """
        Check if a given 32-bit integer key has been set.
        """
        return self.is_member(var)

    def __del__(self):
        if hasattr(self, "_filter"):
            cast_p1 = self._ffi.cast("void*", self._filter.b)
            cast_p2 = self._ffi.cast("void*", self._filter)
            self._libbloomf.free(cast_p1)
            self._libbloomf.free(cast_p2)


class BloomFilterExtended(BloomFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_input_type(self, var: Union[List[int], int, bytes]):
        if isinstance(var, list):
            # Assuming the list is composed of integer elements.
            # Checking its elements for their type is too much of an expensive procedure.
            data = self._ffi.new("GoUint8[]", var)
            length = len(data)
        elif isinstance(var, int):
            data = self._ffi.new("GoUint8[]", [var])
            length = 1
        elif isinstance(var, bytes):
            data = self._ffi.new("GoUint8[]", list(var))
            length = len(data)
        else:
            raise BloomFilterIncorrectInputException(
                "var must be a list of integers, an integer or a bytes object"
            )

        return data, length

    def add_one_member(self, var: Union[List[int], int, bytes]):
        """
        Add a single key to the filter.
        Examples of keys: serialized Python objects, strings, 64-digit integers, etc.
        """
        data, length = self._check_input_type(var)
        go_slice = self._ffi.new(
            "GoSlice*",
            {
                "data": self._ffi.cast("void*", data),
                "len": length,
                "cap": length,
            },
        )
        self._libbloomf.Add(self._filter, go_slice[0])

    def is_one_member(self, var: Union[List[int], int, bytes]) -> bool:
        """
        Check if a single key has been set.
        Examples of keys: serialized Python objects, strings, 64-digit integers, etc.
        """
        data, length = self._check_input_type(var)
        go_slice = self._ffi.new(
            "GoSlice*",
            {
                "data": self._ffi.cast("void*", data),
                "len": length,
                "cap": length,
            },
        )
        return self._libbloomf.Test(self._filter, go_slice[0])



import os

from setuptools import Extension, find_namespace_packages, setup

with open("README.md") as f:
    long_description = f.read()

version = "3.3.0.1"
circleci_build_number = os.getenv("CIRCLE_BUILD_NUM", "")
if circleci_build_number != "":
    version = f"{version}.dev{circleci_build_number}"

setup(
    name="shaped-bloom-filter",
    version=version,
    author="Shaped Team",
    author_email="support@shaped.ai",
    url="https://github.com/shaped-ai/bloom",
    description="Highly optimized bloom filter in Python with bindings in Go.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["bloom-filter", "shaped-ai"],
    classifiers=[
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
    ],
    packages=find_namespace_packages(where="python"),
    package_dir={"": "python"},
    zip_safe=True,
    install_requires=["cffi==1.15.1"],
    build_golang={"root": "github.com/bits-and-blooms/bloom/v3"},
    ext_modules=[
        Extension(
            name="shaped_bloom_filter.libbloomf",
            sources=["python/shaped_bloom_filter/golib/main.go"],
            library_dirs=["python/shaped_bloom_filter/golib"],
        )
    ],
    package_data={"shaped_bloom_filter.golib": ["*.h"]},
    include_package_data=True,
)
